/* Hypergraph Dualization (Hdual.c) */
/* 2003/9/1 Takeaki Uno */
/* This program is available for only academic use.
   Neither commercial use, modification, nor re-distribution is allowed */

#ifndef _shdual_c_
#define _shdual_c_

#include<stdio.h>

/* #define HDUAL_MINIMUM */

int iterations =0;  

/*  iteration */
/* G->edge[i].end, 0<=i<G->node1_num  has to be 0 */
/* crit[0,...,t] critical set, crit->q[crit->end/2,...,+t] of e */
/* uncov: set of sets which include K */
void SHDUAL_iter(SGRAPH *G, BARRAY *SBA, QUEUE *K, QUEUE *crit, QUEUE *uncov){
  QUEUE jump, *jQ, uncov_, crit_;
  int j, u, i, v, t, jj, tt, ii, vv, *s, d;

  iterations ++;

  if ( uncov->t == 0 ){  /* solution if none includes K */
    if ( SHDUAL_output ( G, SBA, K, uncov ) == 0 ) return;
  }
  QUEUE_init ( &jump, G->node1_num );
  malloc2 ( jQ, QUEUE, G->node1_num, "SHdual", "jQ" );

  if ( K->t == 0 ){ /* if K is empty, v not in the top of uncov can be added to K */
    QUEUE_FE_LOOP_ ( G->edge[uncov->q[0]], i, v ) G->edge[v].s = uncov->q[0];
    for ( v=0 ; v<G->node1_num ; v++ ){
      if ( G->edge[v].s != uncov->q[0] ){
        QUEUE_ins_ ( &jump, v );
        jQ[v].s = jQ[v].t = 0;
        jQ[v].q = NULL;
      }
    }
  } else {
      /* make jump list ( list of items have to be checked <= included in a critical set ) */
    QUEUE_FE_LOOP_ ( *crit, j, u )  /* make jump */
      QUEUE_FE_LOOP_ ( G->edge[u], i, v )
        if ( G->edge[v].end++ == 0 ) QUEUE_ins_ ( &jump, v );
    QUEUE_FE_LOOP_ ( jump, i, v ){
      QUEUE_init ( &jQ[v], G->edge[v].end*3+3 );
      jQ[v].end = G->edge[v].end+1;
      jQ[v].q[jQ[v].end-1] = G->node_end;   /*  loop stopper */
      G->edge[v].end = 0;  /* clear used memory */
    }
    QUEUE_FE_LOOP_ ( *crit, j, u ){  /* make jQ */
      QUEUE_FE_LOOP_ ( G->edge[u], i, v ){
        jQ[v].q[jQ[v].t + jQ[v].end ] = crit->q[j+crit->end];
        jQ[v].q[jQ[v].t + jQ[v].end*2 ] = i;
        QUEUE_ins_ ( &jQ[v], u );
      }
    }
  }
  QUEUE_FE_LOOP_ ( G->edge[uncov->q[0]], i, v ) jQ[v].s = -1; /* put mark */

    /* construct cirt(K\cup u) for each u */
  QUEUE_FE_LOOP_ ( jump, i, v ){
    if ( jQ[v].s == -1 ) goto END1;  /* if v is in the top of uncov, then do not add */
    if ( K->t >0 ){
      jQ[v].q[jQ[v].t+jQ[v].end] = G->node1_num;  /* loop stopper */
      t=0; QUEUE_FE_LOOP_ ( *K, j, u ){
        if ( jQ[v].q[jQ[v].end+t] != u || jQ[v].q[t] >= uncov->q[0] ) goto END1;  /* an element of K has empty critical sets: $t$th crit is critical for q[crit->t+t] */
        do { t++; } while ( jQ[v].q[jQ[v].end+t]==u ); /* skip to next crit */
      }
    }
      /* make crit for K \cup v */
    QUEUE_init ( &crit_, (jQ[v].t+uncov->t)*2 );
    crit_.end = jQ[v].t + uncov->t;
    if ( QUEUE_LENGTH ( jQ[v] ) >0 ){
      QUEUE_FE_LOOP_ ( jQ[v], j, u ){
        crit_.q[crit_.t + crit_.end] = jQ[v].q[j+jQ[v].end];
        QUEUE_ins_ ( &crit_, jQ[v].q[j]);
        QUEUE_rm_ ( &(G->edge[u]), jQ[v].q[j+jQ[v].end*2] );     /* remove v from each critical set <= unnecessary in the descendant iterations */
      }
    }
    
      /* update uncov for K \cup v */
    jj=0; t=crit_.t;
    QUEUE_FE_LOOP_ ( *uncov, j, u ){
      if ( BARRAY_BIT(SBA[u], v) ) {
        uncov->q[jj] = uncov->q[j];
        jj++;
      } else {
        crit_.q[crit_.t + crit_.end] = v;
        QUEUE_ins_ ( &crit_, u ); 
        for ( ii=0 ; ii<G->edge[u].t ; ){
          if ( G->edge[G->edge[u].q[ii]].t == -1 ) QUEUE_rm_(&(G->edge[u]),ii);
          else ii++;
        }
      }
    }
    uncov->t = jj;

    QUEUE_ins_ ( K, v );  /* add v to the current solution K */
    G->edge[v].t = -1;    /* mark: edge[v].t = -1 => v is in K */
      SHDUAL_iter( G, SBA, K, &crit_, uncov );   /* recursive call */
    K->t--;               /* remove v from K */
    G->edge[v].t = 0;     /* delete mark */
    crit_.s = t;
      /* recover v to each critical set */
    for ( j=0 ; j<t ; j++ ){
      u = crit_.q[j];
      s = &(G->edge[u].q[jQ[v].q[j+jQ[v].end*2]]);
      QUEUE_ins_ ( &(G->edge[u]), *s ) ;
      *s = v;
    }
      /* merge critical sets for v to uncov <= recovering uncov  */
    QUEUE_FE_LOOP_ ( crit_, j, u ){
      QUEUE_FE_LOOP_ ( *K, ii, vv ){
        QUEUE_ins_ ( &(G->edge[u]), vv );
      }
    }
    QUEUE_merge_ ( uncov, &crit_ ); /* recover uncov */
    QUEUE_end ( &crit_ );
    END1:;
  }
  
  QUEUE_FE_LOOP_ ( jump, i, v ) QUEUE_end ( &jQ[v] );
  QUEUE_end ( &jump );
  free ( jQ );
}


/* dualization main */
void SHDUAL_main ( SGRAPH *SG ){
  int i, j, v;
  QUEUE K, crit, uncov;
  BARRAY *SBA;

  /* allocate memory */
  QUEUE_init ( &uncov, SG->node_end-SG->node1_num );
  QUEUE_init ( &K, SG->node1_num );
  QUEUE_init ( &crit, 3 );
  crit.end = 1;
  for ( i=0 ; i<SG->node1_num ; i++ ){
    SG->edge[i].end = 0;
    SG->edge[i].s = -1;
    SG->edge[i].t = 0;
  }
  malloc2 ( SBA, BARRAY, SG->node_end, "main", "SBA");
  for ( i=SG->node1_num ; i<SG->node_num ; i++){
    QUEUE_ins_ ( &uncov, i );
    BARRAY_init ( &SBA[i], SG->node1_num );
    BARRAY_reset_interval ( &SBA[i], 0, SG->node1_num-1 );
    QUEUE_FE_LOOP_ ( SG->edge[i], j, v ) BARRAY_SET ( SBA[i], v );
  }

  /* call itearion */
    SHDUAL_iter ( SG, SBA, &K, &crit, &uncov );

    /* free memory */
  for ( i=SG->node1_num ; i<SG->node_num ; i++ ) BARRAY_end ( &SBA[i] );
  free ( SBA );
  QUEUE_end ( &K );
  QUEUE_end ( &crit );
  QUEUE_end ( &uncov );
}

#endif

